##### cppsimpyview.py #####
# written by Michael H. Perrott
# available at www.cppsim.com as part of the CppSim package
# Copyright (c) 2004-2023 by Michael H. Perrott
# This file is disributed under the MIT license

# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in
# all copies of this software.

# IN NO EVENT SHALL MICHAEL H PERROTT BE
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
# DOCUMENTATION, EVEN IF MICHAEL H PERROTT
# HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# MICHAEL H PERROTT SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
# HEREUNDER IS ON AN "AS IS" BASIS, AND MICHAEL H PERROTT HAS NO
# OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.


import tkinter as tk
from tkinter import ttk
from tkinter.messagebox import showinfo

import matplotlib
import matplotlib.style as mplstyle
matplotlib.use('MacOSX')
mplstyle.use('fast')
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)
# Implement the default Matplotlib key bindings.
from matplotlib.backend_bases import key_press_handler
from matplotlib.figure import Figure
#import matplotlib_inline.backend_inline
#from matplotlib.widgets import MultiCursor

#import matplotlib.backends.backend_tkagg as tkagg
#from matplotlib.backend_bases import cursors
from matplotlib.patches import Rectangle
from matplotlib.ticker import FormatStrFormatter
from dataclasses import dataclass,field

import os
import sys
import traceback
import numpy as np
import re
from cppsimpyfunctions import *

cppsimsharedhome = os.getenv("CPPSIMSHAREDHOME")
if cppsimsharedhome != None:
   CPPSIMSHARED_PATH = '%s' % cppsimsharedhome
   CPPSIM_PATH = '%s' % os.getenv("CPPSIMHOME")
else:
   home_dir = os.getenv("HOME")
   CPPSIMSHARED_PATH = '%s/CppSim/CppSimShared' % home_dir
   CPPSIM_PATH = '%s/CppSim' % home_dir
sys.path.append(CPPSIMSHARED_PATH + '/Python')
from cppsimdata import *


@dataclass 
class PlotCommandHistory:
   plot_file_name: str = ''
   plot_command_name: str = ''   
   plot_db: bool = False
   plot_logx: bool = False
   plot_logy: bool = False
   plot_xmin: str = ''
   plot_xmax: str = ''   
   command_list: list[str] = field(default_factory=list)
   cur_command_pointer: int = -1
   max_list_length: int = 100

   def set_plot_file_name(self,plot_file_name_in):
      self.plot_file_name = plot_file_name_in
   def get_plot_file_name(self):
      return(self.plot_file_name)
   def set_plot_command_name(self,plot_command_name_in):
      self.plot_command_name = plot_command_name_in
   def get_plot_command_name(self):
      return(self.plot_command_name)

   def set_plot_db(self):
      self.plot_db = True
   def clear_plot_db(self):
      self.plot_db = False
   def get_plot_db(self):
      return(self.plot_db)
   def set_plot_logx(self):
      self.plot_logx = True
   def clear_plot_logx(self):
      self.plot_logx = False
   def get_plot_logx(self):
      return(self.plot_logx)
   def set_plot_logy(self):
      self.plot_logy = True
   def clear_plot_logy(self):
      self.plot_logy = False
   def get_plot_logy(self):
      return(self.plot_logy)
   def set_plot_xmin(self,xmin_val):
      self.plot_xmin = xmin_val
   def set_plot_xmax(self,xmax_val):
      self.plot_xmax = xmax_val
   def add_entry(self,command_in):
      #print("adding entry: '%s'" % command_in)
      #print("Update:  cur_test_tr_filename = '%s'" % self.cur_test_tr_filename)      
      if self.cur_command_pointer >= 0:
         #print("---> cur_command_list = '%s', pointer = %d" % (self.command_list[self.cur_command_pointer],self.cur_command_pointer))
         if command_in not in self.command_list:
            #print("---> APPEND <---")
            self.command_list.append(command_in)
            if len(self.command_list) > self.max_list_length:
               #print("------> DELETE <------")
               #print("---> len = %d, first = '%s', last = '%s'" % (len(self.command_list),self.command_list[0], self.command_list[-1]))
               del self.command_list[0]
               #print("---> len = %d, first = '%s', last = '%s'" % (len(self.command_list),self.command_list[0], self.command_list[-1]))
            self.cur_command_pointer = len(self.command_list)-1
      else:
         self.cur_command_pointer = 0
         self.command_list.append(command_in)         
   def get_prev_entry(self):
      if self.cur_command_pointer > 0:
         self.cur_command_pointer -= 1
         command_out = self.command_list[self.cur_command_pointer]
      elif self.cur_command_pointer == 0:
         command_out = self.command_list[self.cur_command_pointer]
      else:
         command_out = ''
      return(command_out)
   def get_next_entry(self):
      if len(self.command_list) > 0:
         if self.cur_command_pointer < len(self.command_list)-1:
            self.cur_command_pointer += 1
            command_out = self.command_list[self.cur_command_pointer]
         elif self.cur_command_pointer == len(self.command_list)-1:
            command_out = self.command_list[self.cur_command_pointer]
         else:
            command_out = ''
      else:
         command_out = ''         
      return(command_out)
   def get_cur_entry(self):
      if len(self.command_list) > 0:
         command_out = self.command_list[self.cur_command_pointer]
      else:
         command_out = ''         
      return(command_out)
   
@dataclass 
class PlotControlData:
   state: str = 'Off'
   prev_state: str = 'Off'   
   x0: float = 0
   y0: float = 0
   x1: float = 0
   y1: float = 0
   x0_prev: float = 0
   y0_prev: float = 0
   x1_prev: float = 0
   y1_prev: float = 0
   xmin: float = 0
   xmax: float = 0
   x0_rect: float = 0

   def set_x0_rect(self,x0_rect_in):
      self.x0_rect = x0_rect_in
   def get_x0_rect(self):
      return(self.x0_rect)
   def set_state(self,state_in):
      self.prev_state = self.state
      self.state = state_in;
   def get_state(self):
      return(self.state)
   def set_prev_state(self,state_in):
      self.prev_state = state_in;
   def get_prev_state(self):
      return(self.prev_state)
   def set_xmin(self,xmin_in):
      self.xmin = xmin_in;
   def get_xmin(self):
      return(self.xmin)
   def set_xmax(self,xmax_in):
      self.xmax = xmax_in;
   def get_xmax(self):
      return(self.xmax)
   def set_x0(self,x0_in):
      self.x0_prev = self.x0;
      self.x0 = x0_in;
   def set_x1(self,x1_in):
      #print("---***  x1_in = %f" % x1_in)
      self.x1_prev = self.x1;
      self.x1 = x1_in;
      #print("---***  self.x1_in = %f" % self.x1)      
   def set_y0(self,y0_in):
      self.y0_prev = self.y0;
      self.y0 = y0_in;
   def set_y1(self,y1_in):
      self.y1_prev = self.y1;
      self.y1 = y1_in;
   def get_x0(self):
      return(self.x0)
   def get_x1(self):
      return(self.x1)
   def get_y0(self):
      return(self.y0)
   def get_y1(self):
      return(self.y1)
   def get_x0_prev(self):
      return(self.x0_prev)
   def get_x1_prev(self):
      return(self.x1_prev)
   def get_y0_prev(self):
      return(self.y0_prev)
   def get_y1_prev(self):
      return(self.y1_prev)

   
class CppSimPyViewApp(tk.Tk):
   def __init__(self):
       super().__init__()
       
       ## List of plot functions 
       self.plot_command_list = []
       self.plot_command_list.append([transient])
       self.plot_command_list.append([spectrum,'dB','logx'])
       self.plot_command_list.append([phase_noise,'dB','logx'])

       self.plot_command_history_list = []
       for plot_command in self.plot_command_list:
           self.plot_command_history_list.append(PlotCommandHistory())
       i = 0
       for plot_command_history in self.plot_command_history_list:
           plot_command_name = self.plot_command_list[i][0].__name__
           plot_command_arg_list = self.plot_command_list[i]
           plot_file_name = ''
           self.init_plot_command_history(plot_command_history,plot_command_name,plot_command_arg_list,plot_file_name)
           i += 1
           
       self.data = CppSimData()
       self.cur_simruns_path = ''
       self.cur_test_par_filename = ''
       self.cur_test_tr_filename = ''
       self.simulator_name = 'CppSim'
       self.state_multiplot = PlotControlData()       
       self.state_zoomx_all = PlotControlData()
       self.unit_label_dict = {'atto': 1e-18, 
                               'femto': 1e-15,
                               'pico': 1e-12, 
                               'nano': 1e-9,  
                               'micro': 1e-6, 
                               'milli': 1e-3, 
                               'none': 1,  
                               'kilo': 1e3,   
                               'Mega': 1e6,   
                               'Giga': 1e9,   
                               'Tera': 1e12,  
                               }
       self.x_plot_label = ''
       self.xunit_plot_label = ''       
       self.xunit_plot_label_prefix = ''
       self.xunit_scale_factor = 1
       self.y_plot_label = ''
       self.yunit_plot_label = ''       
       self.yunit_plot_label_prefix = ''             
       self.yunit_scale_factor = 1
       self.plot_command = self.plot_command_list[0][0]
       self.plot_db = False
       self.plot_logx = False
       self.plot_logy = False
       self.plot_xmin = ''
       self.plot_xmax = ''
       self.xmin_full = 0
       self.xmax_full = 0
       self.plot_command_history = self.plot_command_history_list[0]
       self.checkbutton_db_value = tk.IntVar(self)
       self.checkbutton_logx_value = tk.IntVar(self)
       self.checkbutton_logy_value = tk.IntVar(self)
       
       #matplotlib_inline.backend_inline.set_matplotlib_formats('retina')
       self.title("CppSimPyView    (Waveform Viewer for CppSim C++ Behavioral Simulator - Written by Michael H. Perrott)")
       matplotlib.rcParams.update({'font.size': 8})
       matplotlib.rcParams.update({'axes.xmargin': 0})               
       self.fig = Figure(figsize=(9,5),dpi=150)
       self.ax_list = []
       self.plot_marker_style = None

       self.style = ttk.Style()
       self.style.configure("mystyle.Treeview", highlightthickness=0, bd=0, font=('Helvetica', 14), background="white", fieldbackground="white",foreground="black") # Modify the font of the body
       self.style.configure("mystyle.Treeview.Heading", font=('Helvetica', 14,'bold'), anchor=tk.W) # Modify the font of the headings
       self.style.layout("mystyle.Treeview", [('mystyle.Treeview.treearea', {'sticky': 'nswe'})]) # Remove the borders

       ############### Top Frame
       self.frame_top = tk.Frame(self,padx=2,pady=2)
       self.frame_top.pack(side=tk.TOP,fill=tk.BOTH,expand=tk.YES)

       ########## Left Frame and Left Sub-Frames
       self.frame_left = tk.Frame(self.frame_top,padx=2,pady=2)
       #self.frame_left = tk.PanedWindow(self.frame_top,showhandle=True,handlepad=10,handlesize=20,orient=tk.VERTICAL)
       self.frame_left.pack(side=tk.LEFT,fill=tk.BOTH)

       ## Buttons for Synch and Load
       self.frame_left_top_buttons = tk.Frame(self.frame_left,padx=1,pady=1)
       self.button_synch = tk.Button(master=self.frame_left_top_buttons, text="Synch", command=self.synch_to_cur_simruns_dir)
       self.button_load = tk.Button(master=self.frame_left_top_buttons, text="Load", command=self.load_data)
       # Pack
       self.frame_left_top_buttons.pack(side=tk.TOP,fill=tk.BOTH)
       self.button_synch.pack(side=tk.LEFT)
       self.button_load.pack(side=tk.LEFT)

       ## Listbox for Par Files
       self.listbox_test_par_names_label = tk.Label(self.frame_left,text='Par File Selection',font=('Helvetica',14,'bold'),anchor='w')
       self.frame_left_listbox_test_par = tk.Frame(self.frame_left,padx=2,pady=2)
       self.listbox_test_par_names = tk.Listbox(self.frame_left_listbox_test_par,width=30,height=2,exportselection=False)
       self.listbox_test_par_names_scrollbar = tk.Scrollbar(self.frame_left_listbox_test_par,command=self.listbox_test_par_names.yview)
       self.listbox_test_par_names.config(yscrollcommand=self.listbox_test_par_names_scrollbar.set)
       self.listbox_test_par_names.bind('<<ListboxSelect>>', self.update_listbox_test_par_names) 
       # Pack
       self.listbox_test_par_names_label.pack(side=tk.TOP,fill=tk.BOTH)
       self.frame_left_listbox_test_par.pack(side=tk.TOP,fill=tk.BOTH)
       self.listbox_test_par_names.pack(side=tk.LEFT,fill=tk.BOTH)
       self.listbox_test_par_names_scrollbar.pack(side=tk.LEFT,fill=tk.BOTH)

       ## Listbox for Tr Files
       self.listbox_test_tr_names_label = tk.Label(self.frame_left,text='Tr File Selection',font=('Helvetica',14,'bold'),anchor='w')
       self.frame_left_listbox_test_tr = tk.Frame(self.frame_left,padx=2,pady=2)
       self.listbox_test_tr_names = tk.Listbox(self.frame_left_listbox_test_tr,width=30,height=3,exportselection=False)
       self.listbox_test_tr_names_scrollbar = tk.Scrollbar(self.frame_left_listbox_test_tr,command=self.listbox_test_tr_names.yview)
       self.listbox_test_tr_names.config(yscrollcommand=self.listbox_test_tr_names_scrollbar.set)
       self.listbox_test_tr_names.bind('<<ListboxSelect>>', self.update_listbox_test_tr_names) 
       # Pack
       self.listbox_test_tr_names_label.pack(side=tk.TOP,fill=tk.BOTH)
       self.frame_left_listbox_test_tr.pack(side=tk.TOP,fill=tk.BOTH)
       self.listbox_test_tr_names.pack(side=tk.LEFT,fill=tk.BOTH)
       self.listbox_test_tr_names_scrollbar.pack(side=tk.LEFT,fill=tk.BOTH)

       ## TreeView for Hierarchical View of Instances
       self.frame_left_tree = tk.Frame(self.frame_left,padx=2,pady=2)
       self.tree_hier_instances = ttk.Treeview(self.frame_left_tree,style="mystyle.Treeview",selectmode="browse",height=4)
       self.tree_hier_instances_scrollbar = ttk.Scrollbar(self.frame_left_tree,command=self.tree_hier_instances.yview)
       self.tree_hier_instances.config(yscrollcommand=self.tree_hier_instances_scrollbar.set)
       self.tree_hier_instances.column('#0',anchor=tk.W, stretch=tk.NO, width=270)
       self.tree_hier_instances.heading('#0', text='Instance Hierarchy                                  ',anchor=tk.W) # Anchor does not seem to work
       self.tree_hier_instances.bind('<Button-1>', self.update_tree_hier_instances)
       # Pack - implement tree approach for node names at a later time
       #self.frame_left_tree.pack(side=tk.TOP,fill=tk.BOTH)
       #self.tree_hier_instances.pack(side=tk.LEFT,fill=tk.BOTH)
       #self.tree_hier_instances_scrollbar.pack(side=tk.LEFT,fill=tk.BOTH)

       ## Listbox for Node Names
       self.listbox_node_names_label = tk.Label(self.frame_left,text='Instance Node Names',font=('Helvetica',14,'bold'),anchor='w')
       self.frame_left_listbox_nodes = tk.Frame(self.frame_left,padx=2,pady=2)
       self.listbox_node_names = tk.Listbox(self.frame_left_listbox_nodes,width=30)
       self.listbox_node_names_scrollbar = tk.Scrollbar(self.frame_left_listbox_nodes,command=self.listbox_node_names.yview)
       self.listbox_node_names.config(yscrollcommand=self.listbox_node_names_scrollbar.set)
       self.listbox_node_names.bind('<Double-1>', self.update_listbox_node_names) 
       # Pack
       self.listbox_node_names_label.pack(side=tk.TOP,fill=tk.BOTH)
       self.frame_left_listbox_nodes.pack(side=tk.TOP,fill=tk.BOTH,expand=tk.YES)
       self.listbox_node_names.pack(side=tk.LEFT,fill=tk.BOTH,expand=tk.YES)
       self.listbox_node_names_scrollbar.pack(side=tk.LEFT,fill=tk.BOTH,expand=tk.YES)

       ## Listbox for Plot Functions
       self.listbox_plot_names_label = tk.Label(self.frame_left,text='Plot Functions',font=('Helvetica',14,'bold'),anchor='w')
       self.frame_left_listbox_plot = tk.Frame(self.frame_left,padx=2,pady=2)
       self.listbox_plot_names = tk.Listbox(self.frame_left_listbox_plot,width=30,height=5,exportselection=False)
       self.listbox_plot_names_scrollbar = tk.Scrollbar(self.frame_left_listbox_plot,command=self.listbox_plot_names.yview)
       self.listbox_plot_names.config(yscrollcommand=self.listbox_plot_names_scrollbar.set)
       self.listbox_plot_names.bind('<<ListboxSelect>>', self.update_listbox_plot_names) 
       # Pack
       self.listbox_plot_names_label.pack(side=tk.TOP,fill=tk.BOTH)
       self.frame_left_listbox_plot.pack(side=tk.TOP,fill=tk.BOTH)
       self.listbox_plot_names.pack(side=tk.LEFT,fill=tk.BOTH)
       self.listbox_plot_names_scrollbar.pack(side=tk.LEFT,fill=tk.BOTH)

       ## Buttons for Plot Commmand
       self.frame_left_bottom_buttons = tk.Frame(self.frame_left,padx=2,pady=6)
       self.button_clear_command_text = tk.Button(master=self.frame_left_bottom_buttons, text="Clear", command=self.clear_text_command)
       self.button_back_command_text = tk.Button(master=self.frame_left_bottom_buttons, text="Prev", command=self.prev_text_command)
       self.button_forward_command_text = tk.Button(master=self.frame_left_bottom_buttons, text="Next", command=self.next_text_command)
       # Pack
       self.frame_left_bottom_buttons.pack(side=tk.TOP,fill=tk.BOTH)
       self.button_clear_command_text.pack(side=tk.LEFT)
       self.button_back_command_text.pack(side=tk.LEFT)
       self.button_forward_command_text.pack(side=tk.LEFT)

       ########## Right Frame and Right Sub-Frames
       ## Right Frame
       self.frame_right = tk.Frame(self.frame_top,padx=0,pady=0)
       self.frame_right.pack(side=tk.RIGHT,fill=tk.BOTH,expand=tk.YES)

       ## Buttons In Right,Top Frame
       self.frame_top_buttons = tk.Frame(self.frame_right,padx=2,pady=2)
       self.button_zoomx = tk.Button(master=self.frame_top_buttons, text="(Z)oom X", command=self.start_zoomx_all)
       self.button_fullx = tk.Button(master=self.frame_top_buttons, text="(F)ull X", command=self.apply_full_zoom_out)
       self.button_zoom_prev = tk.Button(master=self.frame_top_buttons, text="(P)rev Zoom", command=self.apply_prev_zoomx_all)
       self.button_leftpan = tk.Button(master=self.frame_top_buttons, text="<", command=self.apply_pan_left0p1x_all)
       self.button_rightpan = tk.Button(master=self.frame_top_buttons, text=">", command=self.apply_pan_right0p1x_all)
       self.button_zoom_in = tk.Button(master=self.frame_top_buttons, text="In", command=self.apply_zoom2x_in_all)
       self.button_zoom_out = tk.Button(master=self.frame_top_buttons, text="Out", command=self.apply_zoom2x_out_all)
       self.button_measure = tk.Button(master=self.frame_top_buttons, text="(M)easure", command=self.no_op)
       self.button_measure_diff = tk.Button(master=self.frame_top_buttons, text="Meas (D)iff", command=self.no_op)
       self.button_linestyle = tk.Button(master=self.frame_top_buttons, text="(L)inestyle", command=self.apply_toggle_plot_marker_style)
       self.button_save_fig_to_pdf = tk.Button(master=self.frame_top_buttons, text="PDF", command=self.save_plot_to_pdf_file)
       #self.button_save_fig_to_clipboard = tk.Button(master=self.frame_top_buttons, text="Clipboard", command=self.no_op)       
       # Pack
       self.frame_top_buttons.pack(side=tk.TOP,fill=tk.BOTH)
       self.button_zoomx.pack(side=tk.LEFT)
       self.button_fullx.pack(side=tk.LEFT)
       self.button_zoom_prev.pack(side=tk.LEFT)
       self.button_leftpan.pack(side=tk.LEFT)
       self.button_rightpan.pack(side=tk.LEFT)
       self.button_zoom_in.pack(side=tk.LEFT)
       self.button_zoom_out.pack(side=tk.LEFT)
       self.button_measure.pack(side=tk.LEFT)
       self.button_measure_diff.pack(side=tk.LEFT)
       self.button_linestyle.pack(side=tk.LEFT)
       self.button_save_fig_to_pdf.pack(side=tk.LEFT)
       #self.button_save_fig_to_clipboard.pack(side=tk.LEFT)

       ## Buttons in Right,Bottom Frame
       self.frame_bottom_buttons = tk.Frame(self.frame_right,padx=1,pady=1)
       self.button_plot = tk.Button(master=self.frame_bottom_buttons, text="Plot", command=self.plot)
       self.button_load_and_replot = tk.Button(master=self.frame_bottom_buttons, text="Load+Plot", command=self.load_data_and_replot)
       self.checkbutton_db = tk.Checkbutton(master=self.frame_bottom_buttons, text='dB',justify=tk.LEFT,font=('Helvetica',16),
                                            variable=self.checkbutton_db_value,command=self.update_checkbutton_db_value,onvalue=1,offvalue=0,width=4)
       self.checkbutton_logx = tk.Checkbutton(master=self.frame_bottom_buttons, text='logX',justify=tk.LEFT,font=('Helvetica',16),
                                              variable=self.checkbutton_logx_value,command=self.update_checkbutton_logx_value,onvalue=1,offvalue=0,width=5)
       self.checkbutton_logy = tk.Checkbutton(master=self.frame_bottom_buttons, text='logY',justify=tk.LEFT,font=('Helvetica',16),
                                              variable=self.checkbutton_logy_value,command=self.update_checkbutton_logy_value,onvalue=1,offvalue=0,width=5)

       self.entry_xmin_update = tk.Entry(self.frame_bottom_buttons, font=("Arial",14),width=16)
       self.entry_xmin_update.bind("<Return>",self.eval_xmin_update)
       self.label_xmin_update = tk.Label(master=self.frame_bottom_buttons,text='Xmin',font=("Arial",16))
       self.entry_xmax_update = tk.Entry(self.frame_bottom_buttons, font=("Arial",14),width=16)
       self.entry_xmax_update.bind("<Return>",self.eval_xmax_update)
       self.label_xmax_update = tk.Label(master=self.frame_bottom_buttons,text='Xmax',font=("Arial",16))

       # Pack
       self.frame_bottom_buttons.pack(side=tk.BOTTOM,fill=tk.BOTH)
       self.button_plot.pack(side=tk.LEFT)
       self.button_load_and_replot.pack(side=tk.LEFT)
       self.checkbutton_db.pack(side=tk.LEFT)
       self.checkbutton_logx.pack(side=tk.LEFT)
       self.checkbutton_logy.pack(side=tk.LEFT)
       self.entry_xmin_update.pack(side=tk.LEFT)
       self.label_xmin_update.pack(side=tk.LEFT)
       self.entry_xmax_update.pack(side=tk.LEFT)
       self.label_xmax_update.pack(side=tk.LEFT)

       ## Slider in Right, Bottom Frame
       self.slider_update = tk.Scale(self.frame_bottom_buttons, from_=1, to=5, orient=tk.HORIZONTAL,
                              command=self.update_frequency, label="Frequency [Hz]")
       # Pack
       #self.slider_update.pack(side=tk.LEFT)

       ## Canvas for Matplotlib Window and Navigation Toolbar
       self.canvas = FigureCanvasTkAgg(self.fig, master=self.frame_right)  # A tk.DrawingArea.
       self.canvas.draw()
       #self.canvas.mpl_connect("key_press_event", lambda event: print(f"you pressed {event.key}"))
       #self.canvas.mpl_connect("key_press_event", key_press_handler)
       self.canvas.mpl_connect("key_press_event", self.plot_key_press_event)
       self.canvas.mpl_connect("button_press_event", self.plot_button_press_event)
       self.canvas.mpl_connect("button_release_event", self.plot_button_release_event)
       self.canvas.mpl_connect("motion_notify_event", self.plot_motion_notify_event)
       self.canvas.mpl_connect("figure_enter_event", self.plot_figure_enter_event)
       self.canvas.mpl_connect("figure_leave_event", self.plot_figure_leave_event)                
       # Pack
       self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)

       ## Navigation Toolbox for Matplotlib
       self.toolbar = NavigationToolbar2Tk(self.canvas, self.frame_bottom_buttons, pack_toolbar=False)
       #self.toolbar.config(background='red')
       #self.toolbar._message_label.config(background='green')
       self.toolbar._message_label.config(foreground='black')
       #self.toolbar._message_label.config(font=('Helvetica',16,'bold'))
       self.toolbar._message_label.config(font=('Helvetica',16)) 
       self.toolbar.update()
       # Pack
       self.toolbar.pack(side=tk.LEFT, fill=tk.X)

       ############### Bottom Frame with Command Text Line
       self.frame_bottom = tk.Frame(self,padx=0,pady=0,highlightbackground="black",highlightthickness=1)
       self.text_command = tk.Entry(self.frame_bottom, font=("Arial",17))
       self.text_command.bind("<Return>",self.eval_text_command)
       # Pack
       self.frame_bottom.pack(side=tk.BOTTOM,padx=5,pady=5,fill=tk.BOTH,expand=tk.NO)
       self.text_command.pack(side=tk.BOTTOM, fill=tk.X, expand=tk.YES)

       #################
       self.fill_tree_hier_instances()
       self.fill_listbox_plot_names()
       #################

   def set_entry_xmin_str(self,xmin_string):
       if xmin_string == '':
          self.entry_xmin_update.delete(0,tk.END)
       else:
          self.entry_xmin_update.delete(0,tk.END)
          self.entry_xmin_update.insert(0,xmin_string)             

   def set_entry_xmax_str(self,xmax_string):
       if xmax_string == '':
          self.entry_xmax_update.delete(0,tk.END)
       else:
          self.entry_xmax_update.delete(0,tk.END)
          self.entry_xmax_update.insert(0,xmax_string)             

   def set_entry_xmin_val(self,xmin_val):
       xmin_string = '%5.9e' % xmin_val
       self.entry_xmin_update.delete(0,tk.END)
       self.entry_xmin_update.insert(0,xmin_string)             

   def set_entry_xmax_val(self,xmax_val):
       xmax_string = '%5.9e' % xmax_val
       self.entry_xmax_update.delete(0,tk.END)
       self.entry_xmax_update.insert(0,xmax_string)             
       
   def init_plot_command_history(self,plot_command_history,plot_command_name,plot_command_arg_list,plot_file_name):
      plot_command_history.set_plot_command_name(plot_command_name)
      plot_command_history.set_plot_file_name(plot_file_name)      
      plot_command_history.clear_plot_db()
      plot_command_history.clear_plot_logx()
      plot_command_history.clear_plot_logy()
      plot_command_history.set_plot_xmin('')
      plot_command_history.set_plot_xmax('')           
      for j in range(len(plot_command_arg_list)):
         if j > 0:
            plot_command_item = plot_command_arg_list[j]
            if plot_command_item == 'dB':
               plot_command_history.set_plot_db()
            elif plot_command_item == 'logx':
               plot_command_history.set_plot_logx()
            elif plot_command_item == 'logy':
               plot_command_history.set_plot_logy()
            elif 'xmin=' in plot_command_item:
               xmin_extract = plot_command_item.split('=')
               plot_command_history.set_plot_xmin(xmin_extract[1])
            elif 'xmax=' in plot_command_item:
               xmax_extract = plot_command_item.split('=')
               plot_command_history.set_plot_xmax(xmax_extract[1])                    
       
   def update_checkbutton_db_value(self):
       self.modify_listbox_plot_names()
       self.plot_text_command()

   def update_checkbutton_logx_value(self):
       self.modify_listbox_plot_names()      
       self.plot_text_command()

   def update_checkbutton_logy_value(self):
       self.modify_listbox_plot_names()      
       self.plot_text_command()

   def no_op(self):
       pass

   #@profile
   def determine_unit_prefix(self,x_in):
       #x_in_max = max(x_in)
       x_in_max = x_in[-1] ## assume that x axis is monotonically increasing
       dict_reduced = {k:v for (k,v) in self.unit_label_dict.items() if v <  x_in_max}
       unit_label = max(dict_reduced,key=dict_reduced.get)
       #unit_label_list = list(self.unit_label_dict.keys())
       #unit_value_list = list(self.unit_label_dict.values())
       #unit_label = unit_label_list[int(np.argwhere(unit_value_list < max(x_in))[-1])]
       #print("unit_label = '%s'" % unit_label)
       unit_factor = self.unit_label_dict[unit_label]
       if abs(unit_factor) > 1e-30:
           scale_factor = 1/unit_factor
       else:
           scale_factor = 1
       if unit_label == 'default':
          unit_label = ''
       return(unit_label,scale_factor)
       
   def save_plot_to_pdf_file(self):
       print_directory = '%s/PDF' % CPPSIM_PATH
       if not os.path.isdir(print_directory):
          try:
             os.makedirs(print_directory)
          except OSError as error:
             print("Error in 'save_plot_to_pdf_file:")
             print("   --> dir '%s' could not be created" % print_directory)
             return

       # Determine filename while avoiding name clash with existing pdf files
       filename_list = os.listdir(print_directory)       
       i = 0
       while(1):
          i += 1
          print_filename = 'cppsim_plot%d.pdf' % i
          not_found_flag = 0
          for filename in filename_list:
              #print("filename = '%s', print_filename = '%s'" % (filename,print_filename))
              if print_filename in filename:
                  not_found_flag = 1
                  break
          if not_found_flag == 0:
             break;
       full_print_filename = '%s/%s' % (print_directory,print_filename)
       self.fig.savefig(full_print_filename)
       tk.messagebox.showinfo(title='PDF Generated',message="File location: '%s'"%full_print_filename)

   def apply_toggle_plot_marker_style(self):
       if self.plot_marker_style == None:
          self.plot_marker_style = 'o'
       else:
          self.plot_marker_style = None
       self.plot_text_command()

       
   def apply_full_zoom_out(self):
       self.state_multiplot.set_state('Off')
       #for ax in self.ax_list:
       #   xmin,xmax = ax.get_xlim()       
       #   break
       #self.state_multiplot.set_x0(xmin) 
       #self.state_multiplot.set_x1(xmax)
       self.state_multiplot.set_x0(self.xmin_full) 
       self.state_multiplot.set_x1(self.xmax_full)       
       self.set_entry_xmin_str('') 
       self.set_entry_xmax_str('') 
       self.modify_listbox_plot_names()       
       self.plot_text_command()

   def apply_zoom2x_in_all(self):
       if self.state_multiplot.get_state() == 'Off':
          self.state_multiplot.set_state('ZoomXall')
          xmin = self.xmin_full
          xmax = self.xmax_full
       else:
          xmin = self.state_multiplot.get_x0() 
          xmax = self.state_multiplot.get_x1()
       xdelta = (xmax-xmin)/4
       x0_new = xmin + xdelta
       x1_new = xmax - xdelta
       self.state_multiplot.set_x0(x0_new)
       self.state_multiplot.set_x1(x1_new)
       self.set_entry_xmin_val(x0_new) 
       self.set_entry_xmax_val(x1_new) 
       self.modify_listbox_plot_names()              
       self.plot_text_command()

   def apply_zoom2x_out_all(self):
       if self.state_multiplot.get_state() == 'ZoomXall':
          xmin = self.state_multiplot.get_x0() 
          xmax = self.state_multiplot.get_x1()
          xdelta = (xmax-xmin)/2
          x0_new = xmin-xdelta
          x1_new = xmax+xdelta          
          if x0_new < self.state_multiplot.get_xmin():
             x0_new = self.state_multiplot.get_xmin()
          if x1_new > self.state_multiplot.get_xmax():
             x1_new = self.state_multiplot.get_xmax()

          if x0_new <= self.xmin_full*(1+1e-3) and x1_new >= self.xmax_full*(1-1e-3):
             x0_new = self.xmin_full
             x1_new = self.xmax_full
             self.set_entry_xmin_str('') 
             self.set_entry_xmax_str('') 
             self.state_multiplot.set_state('Off')
          else:
             self.set_entry_xmin_val(x0_new) 
             self.set_entry_xmax_val(x1_new)

          self.state_multiplot.set_x0(x0_new) 
          self.state_multiplot.set_x1(x1_new)             
          self.modify_listbox_plot_names()              
          self.plot_text_command()

   def apply_pan_right0p1x_all(self):
       if self.state_multiplot.get_state() == 'ZoomXall':
          xmin = self.state_multiplot.get_x0() 
          xmax = self.state_multiplot.get_x1()
          xdelta = (xmax-xmin)/10
          x1_new = xmax+xdelta
          if x1_new > self.state_multiplot.get_xmax():
             xdelta = self.state_multiplot.get_xmax() - xmax
          x0_new = xmin+xdelta
          x1_new = xmax+xdelta          
          self.state_multiplot.set_x0(x0_new) 
          self.state_multiplot.set_x1(x1_new)
          self.set_entry_xmin_val(x0_new) 
          self.set_entry_xmax_val(x1_new) 
          self.modify_listbox_plot_names()  
          self.plot_text_command()

   def apply_pan_left0p1x_all(self):
       if self.state_multiplot.get_state() == 'ZoomXall':
          xmin = self.state_multiplot.get_x0() 
          xmax = self.state_multiplot.get_x1()
          xdelta = (xmax-xmin)/10
          x0_new = xmin-xdelta
          if x0_new < self.state_multiplot.get_xmin():
             xdelta = xmin - self.state_multiplot.get_xmin()
          x0_new = xmin-xdelta
          x1_new = xmax-xdelta             
          self.state_multiplot.set_x0(x0_new) 
          self.state_multiplot.set_x1(x1_new)
          self.set_entry_xmin_val(x0_new) 
          self.set_entry_xmax_val(x1_new) 
          self.modify_listbox_plot_names()  
          self.plot_text_command()
          
   def apply_prev_zoomx_all(self):
       cur_state = self.state_multiplot.get_state()
       prev_state = self.state_multiplot.get_prev_state()
       self.state_multiplot.set_prev_state(cur_state)       
       if prev_state == 'ZoomXall':
          x0_new = self.state_multiplot.get_x0_prev()  
          x1_new = self.state_multiplot.get_x1_prev()        
          self.state_multiplot.set_x0(x0_new)
          self.state_multiplot.set_x1(x1_new)
          self.set_entry_xmin_val(x0_new) 
          self.set_entry_xmax_val(x1_new)
          self.state_multiplot.set_state('ZoomXall') 
       elif prev_state == 'Off':
          x0_new = self.xmin_full
          x1_new = self.xmax_full
          self.set_entry_xmin_str('') 
          self.set_entry_xmax_str('')
          self.state_multiplot.set_state('Off')     
       self.modify_listbox_plot_names()            
       self.plot_text_command()
       
   def plot_key_press_event(self,event):
       #print("you pressed key")
       #print(event.key)
       if event.key == 'z' or event.key == 'Z':
          self.start_zoomx_all()
       elif event.key == 'f' or event.key == 'F': 
          self.apply_full_zoom_out()
       elif event.key == 'p' or event.key == 'P': 
          self.apply_prev_zoomx_all()
       elif event.key == 'left':
          self.apply_pan_left0p1x_all()
       elif event.key == 'right':
          self.apply_pan_right0p1x_all()
       elif event.key == 'down':
          self.apply_zoom2x_in_all()
       elif event.key == 'up':          
          self.apply_zoom2x_out_all()
       elif event.key == 'l' or event.key == 'L': 
          self.apply_toggle_plot_marker_style()
          
       #print("-----------------")

   #@profile
   def start_zoomx_all(self):
       if len(self.ax_list) > 0:
          self.state_zoomx_all.set_state('First')
          fig_width,fig_height = self.fig.get_size_inches()*self.fig.dpi
          xmin_fig = 0
          xmax_fig = fig_width
          ymin_fig = 0
          ymax_fig = fig_height
          x0_rect = xmin_fig+(xmax_fig-xmin_fig)/10
          y0_rect = 0
          width_rect = (xmax_fig-xmin_fig)/200
          height_rect = ymax_fig-ymin_fig
          found_patch_flag = False
          for patch in self.fig.patches:
              if patch.get_label() == 'rect_zoomx_all':
                  patch.set(visible=True)
                  found_patch_flag = True
                  break
          if found_patch_flag == False:
             rect = Rectangle((x0_rect,y0_rect),width_rect,height_rect,label='rect_zoomx_all')
             #rect.set(fill=True,facecolor='black',alpha=0.2,transform=self.fig.transFigure,figure=self.fig)
             rect.set(fill=True,facecolor='black',alpha=0.2,figure=self.fig)       
             self.fig.patches.extend([rect])
          self.canvas.draw()

          # Save data coordinates
          xmin,xmax = self.ax_list[0].get_xlim()       
          ymin,ymax = self.ax_list[0].get_ylim()
          self.state_multiplot.set_x0(xmin)
          self.state_multiplot.set_x1(xmax)

   def plot_button_press_event(self,event):
       #print("##### you pressed button")
       #print(event.button,event.inaxes,event.xdata,event.ydata,event.x,event.y)
       #print("------------------")
       if self.state_zoomx_all.get_state() == 'First':
           self.state_zoomx_all.set_state('Second')
           if len(self.ax_list) > 0:
              ax_val = self.ax_list[-1]
              inv_func = ax_val.transData.inverted()
              xy_val = inv_func.transform((event.x,event.y))
              #print(xy_val)
              xdata_val = xy_val[0]/self.xunit_scale_factor
           else:
              xdata_val = 0
           #print("------> xdata = %d, x = %d ----" % (xdata_val,event.x))
           self.state_zoomx_all.set_x0(xdata_val)
           self.state_zoomx_all.set_x0_rect(event.x)           
           for patch in self.fig.patches:
              if patch.get_label() == 'rect_zoomx_all':
                 patch.set_x(event.x)
           self.canvas.draw()                      
       elif self.state_zoomx_all.get_state() == 'Second':
           self.state_zoomx_all.set_state('Off')
           if len(self.ax_list) > 0:
              ax_val = self.ax_list[-1]
              inv_func = ax_val.transData.inverted()
              xy_val = inv_func.transform((event.x,event.y))
              #print(xy_val)
              xdata_val = xy_val[0]/self.xunit_scale_factor
           else:
              xdata_val = 0           
           #print("------> xdata = %d, x = %d ----" % (xdata_val,event.x))           
           #print("------> x1 from meas = %f" % event.xdata)
           self.state_zoomx_all.set_x1(xdata_val)
           #print("------> x1 from structure = %f" % self.state_zoomx_all.get_x1())           
           for patch in self.fig.patches:
              if patch.get_label() == 'rect_zoomx_all':
                 patch.set(visible=False)
           self.canvas.draw()
           #print("x0 = %f, x1 = %f" % (self.state_zoomx_all.get_x0(),self.state_zoomx_all.get_x1()))
           self.state_multiplot.state = 'ZoomXall'
           # set xmin and xmax Entry values if plot_min or plot_xmax values are not empty (i.e. are not = '')
           plot_xmin = min([self.state_zoomx_all.get_x0(),self.state_zoomx_all.get_x1()])
           plot_xmax = max([self.state_zoomx_all.get_x0(),self.state_zoomx_all.get_x1()])
           #self.set_entry_xmin_val(plot_xmin) zippy
           #self.set_entry_xmax_val(plot_xmax)           
           self.state_multiplot.set_x0(plot_xmin)
           self.state_multiplot.set_x1(plot_xmax)
           self.set_entry_xmin_val(plot_xmin) 
           self.set_entry_xmax_val(plot_xmax) 
           self.modify_listbox_plot_names()  
           self.plot_text_command()
       ## Code below is just to show how to determine subplot number for future use
       sel_ax = -1
       i = 0
       for ax in self.ax_list:
          if ax == event.inaxes:
             sel_ax = i
             break
          i = i + 1
       #print("---> axis number is %d" % sel_ax)
       #print("-----------------")

   def plot_motion_notify_event(self,event):
       #print("you moved mouse")
       #print(event.button,event.inaxes,event.xdata,event.ydata,event.x,event.y)
       #if len(self.ax_list) > 0:
       #   ax_val = self.ax_list[-1]
       #   inv_func = ax_val.transData.inverted()
       #   xy_val = inv_func.transform((event.x,event.y))
       #   print(xy_val)
       if self.state_zoomx_all.get_state() == 'First':
           for patch in self.fig.patches:
              #print(patch)
              #print(patch.get_label())
              if patch.get_label() == 'rect_zoomx_all':
                 patch.set_x(event.x)
           self.fig.canvas.draw() # could also be self.canvas.draw()
       elif self.state_zoomx_all.get_state() == 'Second':
           for patch in self.fig.patches:
              if patch.get_label() == 'rect_zoomx_all':
                 patch.set_width(event.x-self.state_zoomx_all.get_x0_rect())
           self.fig.canvas.draw() # could also be self.canvas.draw()
       #print("-----------------")
        
   def plot_button_press_event_ax(self,event):
       #print("##### you pressed button")
       #print(event.button,event.inaxes,event.xdata,event.ydata,event.x,event.y)
       #print("------------------")
       if self.state_zoomx_all.get_state() == 'First' and event.xdata != None:
           self.state_zoomx_all.set_state('Second')
           self.state_zoomx_all.set_x0(event.xdata)
           for ax in self.ax_list:
               for patch in ax.patches:
                  if patch.get_label() == 'rect_zoomx_all':
                      patch.set_x(event.xdata)
           self.canvas.draw()                      
       elif self.state_zoomx_all.get_state() == 'Second' and event.xdata != None:
           self.state_zoomx_all.set_state('Off')
           #print("------> x1 from meas = %f" % event.xdata)
           self.state_zoomx_all.set_x1(event.xdata)
           #print("------> x1 from structure = %f" % self.state_zoomx_all.get_x1())           
           for ax in self.ax_list:
               for patch in ax.patches:
                  if patch.get_label() == 'rect_zoomx_all':
                      patch.remove()
           self.canvas.draw()
           #print("x0 = %f, x1 = %f" % (self.state_zoomx_all.get_x0(),self.state_zoomx_all.get_x1()))
           self.state_multiplot.state = 'ZoomXall'
           self.state_multiplot.set_x0(min([self.state_zoomx_all.get_x0(),self.state_zoomx_all.get_x1()]))
           self.state_multiplot.set_x1(max([self.state_zoomx_all.get_x0(),self.state_zoomx_all.get_x1()]))
           self.plot_text_command()
       ## Code below is just to show how to determine subplot number for future use           
       sel_ax = -1
       i = 0
       for ax in self.ax_list:
          if ax == event.inaxes:
             sel_ax = i
             break
          i = i + 1
       #print("---> axis number is %d" % sel_ax)
       #print("-----------------")

   def plot_motion_notify_event_ax(self,event): #### Not used since it was too slow with many subplots ####
       #print("you moved mouse")
       #print(event.button,event.inaxes,event.xdata,event.ydata)
       if self.state_zoomx_all.get_state() == 'First' and event.xdata != None:
           for ax in self.ax_list:
               for patch in ax.patches:
                  if patch.get_label() == 'rect_zoomx_all':
                      patch.set_x(event.xdata) 
           self.canvas.draw()
       elif self.state_zoomx_all.get_state() == 'Second' and event.xdata != None:
           for ax in self.ax_list:
               for patch in ax.patches:
                  if patch.get_label() == 'rect_zoomx_all':
                      patch.set_width(event.xdata-self.state_zoomx_all.get_x0())
           self.canvas.draw()
       #print("-----------------")
   ##################################################################################################
       
   def plot_button_release_event(self,event):
       #print("you released button")
       #print(event.button,event.inaxes,event.xdata,event.ydata)       
       #print("-----------------")
       pass

   def plot_figure_enter_event(self,event):
       #print("you entered figure")
       #print(event.button,event.inaxes,event.xdata,event.ydata)              
       self.canvas.get_tk_widget().focus_set()
       #print("-----------------")

   def plot_figure_leave_event(self,event):
       #print("you left figure")
       #print(event.button,event.inaxes,event.xdata,event.ydata)              
       self.listbox_test_par_names_label.focus_set() # need to focus on something, so focus on label that is immune to key and mouse strokes
       #print("-----------------")

   def fill_listbox_test_par_names(self,par_file_names_list):
       self.listbox_test_par_names.delete(0,tk.END)
       if len(par_file_names_list) > 0:
          selected_index = 0
          count = 0
          for par_file_name in par_file_names_list:
             if par_file_name == self.cur_test_par_filename:
                 selected_index = count
             self.listbox_test_par_names.insert(count,par_file_name)
             count = count + 1
          self.listbox_test_par_names.select_set(selected_index)
          self.listbox_test_par_names.event_generate("<<ListboxSelect>>") ## calls self.update_listbox_test_par_names()
       else:
           self.cur_test_par_filename = ''

           
   def update_listbox_test_par_names(self,event):
       index_val = self.listbox_test_par_names.curselection()
       if len(index_val) > 0:
          self.cur_test_par_filename = self.listbox_test_par_names.get(index_val)
          #print("Update:  cur_test_par_filename = '%s'" % self.cur_test_par_filename)
          self.fill_listbox_test_tr_names()
       else:
          self.cur_test_par_filename = ''
          self.cur_test_tr_filename = ''
       self.fill_listbox_node_names(assert_load=False)

   def fill_listbox_test_tr_names(self):
       # Find all core TR filenames from Selected Par file 
       test_tr_file_names_core_list = []
       if len(self.cur_test_par_filename) > 0:
           cur_test_par_filename = "%s/%s" % (self.cur_simruns_path,self.cur_test_par_filename)
           with open(cur_test_par_filename) as fp:
               lines = fp.readlines()
               for line in lines:
                   if 'output:' in line:
                      split_line = re.sub(' +',' ',line).split(' ',2)
                      # print(split_line)
                      if split_line[0] == 'output:':
                         test_tr_file_names_core_list.append(split_line[1])
               fp.close()
           test_tr_file_names_core_list.sort()

       # Find all tr files associated with each core filename
       cur_simruns_path_filelist = os.listdir(self.cur_simruns_path)
       test_tr_file_names_list = []
       for test_tr_file_name in test_tr_file_names_core_list:
           test_tr_file_name_expanded = '%s.tr' % test_tr_file_name
           for filename in cur_simruns_path_filelist:
              if test_tr_file_name_expanded in filename:
                 test_tr_file_names_list.append(filename)
       test_tr_file_names_list.sort()
       #print(test_tr_file_names_list)

       # Update Tr File Names Listbox
       self.listbox_test_tr_names.delete(0,tk.END)
       if len(test_tr_file_names_list) > 0:
           selected_index = 0
           count = 0
           for test_tr_file_name in test_tr_file_names_list:
               if test_tr_file_name == self.cur_test_tr_filename:
                   selected_index = count
               self.listbox_test_tr_names.insert(count,test_tr_file_name)
               count = count + 1
           self.listbox_test_tr_names.select_set(selected_index)
           self.listbox_test_tr_names.event_generate("<<ListboxSelect>>") ## calls self.update_listbox_test_tr_names()
       else:
           self.cur_test_tr_filename = ''

   def update_listbox_test_tr_names(self,event):
       index_val = self.listbox_test_tr_names.curselection()
       if len(index_val) > 0:
          self.cur_test_tr_filename = self.listbox_test_tr_names.get(index_val)
          #print("Update:  cur_test_tr_filename = '%s'" % self.cur_test_tr_filename)
       else:
          self.cur_test_tr_filename = ' '
       self.update_plot_command_history()
       updated_command_string = self.plot_command_history.get_cur_entry()
       self.text_command.delete(0,tk.END)
       self.text_command.insert(0,updated_command_string)
       if len(updated_command_string) > 0:
          self.fill_listbox_node_names(assert_load=True)
          self.plot_text_command()
       else:
          self.fill_listbox_node_names(assert_load=False)

   def fill_listbox_node_names(self,assert_load=True):
       full_cur_test_tr_filename = '%s/%s' % (self.cur_simruns_path,self.cur_test_tr_filename)
       if len(self.cur_test_tr_filename) > 0 and assert_load == True:
            node_names_list = []
            cur_test_tr_filename_full = '%s/%s' % (self.cur_simruns_path,self.cur_test_tr_filename)
            if os.path.exists(cur_test_tr_filename_full.encode('utf-8')):
               file_stats = os.stat(cur_test_tr_filename_full.encode('utf-8'))
               #print('File size = %d' % file_stats.st_size)
               if file_stats.st_size > 0:
                  #print("---> start loadsig")
                  self.data.loadsig(cur_test_tr_filename_full.encode('utf-8'))
                  #print("---> end loadsig")
                  node_names_list = self.data.lssig()
                  node_names_list.sort()
       elif self.data.get_filename() == full_cur_test_tr_filename.encode('utf-8'):
            node_names_list = self.data.lssig()
            node_names_list.sort()            
       else:
            node_names_list = []

       self.listbox_node_names.delete(0,tk.END)
       count = 0
       if len(node_names_list) > 0:
            for node_name in node_names_list:
               self.listbox_node_names.insert(count,node_name)
               count = count + 1

   def prev_text_command(self):           
       updated_command_string = self.plot_command_history.get_prev_entry()
       self.text_command.delete(0,tk.END)
       self.text_command.insert(0,updated_command_string)          

   def next_text_command(self):           
       updated_command_string = self.plot_command_history.get_next_entry()
       self.text_command.delete(0,tk.END)
       self.text_command.insert(0,updated_command_string)          
       
   def clear_text_command(self):
       self.text_command.delete(0,tk.END)
       
   def update_listbox_node_names(self,event):
       index_val = self.listbox_node_names.curselection()
       if len(index_val) > 0:
          node_to_add = self.listbox_node_names.get(index_val)          
          command_string = self.text_command.get()
          if len(command_string) > 0:
             updated_command_string = command_string.replace('#',node_to_add,1)
             if updated_command_string == command_string:
                updated_command_string += ';%s' % node_to_add
          else:
             updated_command_string = '%s' % node_to_add
          self.text_command.delete(0,tk.END)
          self.text_command.insert(0,updated_command_string)          
          if '#' not in updated_command_string:
             self.plot_text_command()

   def fill_listbox_plot_names(self):
       i = 0
       for plot_command in self.plot_command_list:
          listbox_plot_command_string = plot_command[0].__name__
          for j in range(len(plot_command)):
             if j > 0:
                listbox_plot_command_string += ',' + plot_command[j]
          self.listbox_plot_names.insert(i,listbox_plot_command_string)
          i += 1
       selected_index = 0
       self.listbox_plot_names.select_set(selected_index)
       self.listbox_plot_names.event_generate("<<ListboxSelect>>") ## calls self.update_listbox_plot_names()


   def modify_listbox_plot_names(self):
       index_val = self.listbox_plot_names.curselection()          
       index_val_int = index_val[0]
       plot_string_list = self.listbox_plot_names.get(index_val_int).split(',')
       self.listbox_plot_names.delete(index_val_int)
       #print(plot_string_list)
       listbox_plot_command_string = plot_string_list[0]
       plot_command_list = [self.plot_command_list[index_val_int][0]]
       if self.checkbutton_db_value.get() == 1:
          listbox_plot_command_string += ',dB'
          self.plot_db = True
          plot_command_list.append('dB')
       else:
          self.plot_db = False
       if self.checkbutton_logx_value.get() == 1:
          listbox_plot_command_string += ',logx'
          self.plot_logx = True
          plot_command_list.append('logx')
       else:
          self.plot_logx = False
       if self.checkbutton_logy_value.get() == 1:
          listbox_plot_command_string += ',logy'
          self.plot_logy = True
          plot_command_list.append('logy')
       else:
          self.plot_logy = False
       xmin_string = self.entry_xmin_update.get()
       if xmin_string != '':
          listbox_plot_command_string += ',xmin=' + xmin_string
          self.plot_xmin = xmin_string
          plot_command_list.append('xmin=%s'%xmin_string)
       else:
          self.plot_xmin = ''
       xmax_string = self.entry_xmax_update.get()
       if xmax_string != '':
          listbox_plot_command_string += ',xmax=' + xmax_string
          self.plot_xmax = xmax_string
          plot_command_list.append('xmax=%s'%xmax_string)          
       else:
          self.plot_xmax = ''
       self.plot_command_list[index_val_int] = plot_command_list
       self.listbox_plot_names.insert(index_val_int,listbox_plot_command_string)
       self.listbox_plot_names.select_set(index_val_int)       

   def update_listbox_plot_names(self,event):
       index_val = self.listbox_plot_names.curselection()
       index_val_int = index_val[0]
       if len(index_val) > 0:
          self.plot_command = self.plot_command_list[index_val_int][0]
          self.plot_db = False
          self.plot_logx = False
          self.plot_logy = False
          self.plot_xmin = ''
          self.plot_xmax = ''          
          for j in range(len(self.plot_command_list[index_val_int])):
              if j > 0:
                 plot_command_item = self.plot_command_list[index_val_int][j]
                 if plot_command_item == 'dB':
                    self.plot_db = True
                 elif plot_command_item == 'logx':
                    self.plot_logx = True
                 elif plot_command_item == 'logy':
                    self.plot_logy = True
                 elif 'xmin=' in plot_command_item:
                    xmin_extract = plot_command_item.split('=') 
                    self.plot_xmin = xmin_extract[1]
                 elif 'xmax=' in plot_command_item:
                    xmax_extract = plot_command_item.split('=') 
                    self.plot_xmax = xmax_extract[1]
          if self.plot_db == True:
             self.checkbutton_db.select()
          else:
             self.checkbutton_db.deselect()             
          if self.plot_logx == True:
             self.checkbutton_logx.select()
          else:
             self.checkbutton_logx.deselect()             
          if self.plot_logy == True:
             self.checkbutton_logy.select()
          else:
             self.checkbutton_logy.deselect()

          # set xmin and xmax Entry values if plot_min or plot_xmax values are not empty (i.e. are not = '') 
          self.set_entry_xmin_str(self.plot_xmin)       
          self.set_entry_xmax_str(self.plot_xmax)
          if self.plot_xmin == '':
             self.state_multiplot.set_x0(0) 
          else:
             self.state_multiplot.set_x0(float(self.plot_xmin))             
          if self.plot_xmax == '':
             self.state_multiplot.set_x1(0)
          else:
             self.state_multiplot.set_x1(float(self.plot_xmax))             

          #print("db, logx, logy:")
          #print(self.plot_db)
          #print(self.plot_logx)
          #print(self.plot_logy)
          #print("Update:  cur_test_tr_filename = '%s'" % self.cur_test_tr_filename)

          self.update_plot_command_history()          
          #self.plot_command_history = self.plot_command_history_list[index_val_int]
          updated_command_string = self.plot_command_history.get_cur_entry()
          #print("index = %d, updated command string = '%s'" % (index_val_int,updated_command_string))
          self.text_command.delete(0,tk.END)
          self.text_command.insert(0,updated_command_string)
          self.plot_text_command()

      
   def fill_tree_hier_instances(self):
       instance1 = self.tree_hier_instances.insert('','end',iid=0,text="Item_0",values=("Instance0"))
       instance2 = self.tree_hier_instances.insert(instance1,'end',iid=1,text="Item_1",values=("Instance1"))
       instance3 = self.tree_hier_instances.insert(instance1,'end',iid=2,text="Item_2",values=("Instance2"))
       instance4 = self.tree_hier_instances.insert('','end',iid=3,text="Item_3",values=("Instance3"))
       instance5 = self.tree_hier_instances.insert(instance4,'end',iid=4,text="Item_4",values=("Instance4"))
       instance6 = self.tree_hier_instances.insert(instance4,'end',iid=5,text="Item_5",values=("Instance5"))
       instance7 = self.tree_hier_instances.insert('','end',iid=6,text="Item_6",values=("Instance6"))
       instance8 = self.tree_hier_instances.insert(instance7,'end',iid=7,text="Item_7",values=("Instance7"))
       instance9 = self.tree_hier_instances.insert(instance7,'end',iid=8,text="Item_8",values=("Instance8"))
       instance10 = self.tree_hier_instances.insert('','end',iid=9,text="Item_9",values=("Instance9"))
       instance11 = self.tree_hier_instances.insert(instance10,'end',iid=10,text="Item_10",values=("Instance10"))
       instance12 = self.tree_hier_instances.insert(instance10,'end',iid=11,text="Item_11",values=("Instance11"))

   def update_tree_hier_instances(self,event):
       if self.tree_hier_instances.selection() != ():
           item = self.tree_hier_instances.selection()[0]
           #print("you clicked on", self.tree_hier_instances.item(item,"text"))

   def update_frequency(self,new_val):
       # retrieve frequency
       f = float(new_val)

       # update data
       y = 2 * np.sin(2 * np.pi * f * t)
       self.line1.set_data(t, y)
       self.line2.set_data(t, y)

       # required to update canvas and attached toolbar!
       self.canvas.draw()

   def extract_nodes(self,expression_inp):
       node_name_dict = {}
       full_cur_test_tr_filename = '%s/%s' % (self.cur_simruns_path,self.cur_test_tr_filename)
       #print(self.data.get_filename())
       #print(full_cur_test_tr_filename.encode('utf-8'))
       if self.data.get_filename() == full_cur_test_tr_filename.encode('utf-8'):
          expression_inp_cleaned = expression_inp.replace('.','_')
          node_name_list = self.data.lssig()
          for node_name in node_name_list:
              #print("node name = '%s'" % node_name)
              node_name_cleaned = node_name.replace('.','_')
              if expression_inp_cleaned.find(node_name_cleaned) >= 0:
                  node_sig = self.data.evalsig(node_name)
                  node_name_dict.update({node_name_cleaned: node_sig})
          node_name_dict.update({'__builtins__': {}})
       return(node_name_dict)

   def clean_expression_string(self,expression_inp):
       len_expr = len(expression_inp)
       expression_inp_cleaned = ''
       i = 0;
       for char_val in expression_inp:
          new_char_val = char_val
          if char_val == '.':
             prev_val = '0'
             next_val = '0'
             if i > 0:
                prev_val = expression_inp[i-1]
             if i < (len_expr-1):
                next_val = expression_inp[i+1]
             prev_val_check = ((prev_val >= '0') and (prev_val <= '9')) or \
                               (prev_val == '+') or (prev_val == '-') or (prev_val == '/') or (prev_val == '*') or \
                               (prev_val == ',') or (prev_val == ';')
             next_val_check = ((next_val >= '0') and (next_val <= '9')) or (next_val == 'e') or (next_val <= 'E') or \
                               (next_val == '+') or (next_val == '-') or (next_val == '/') or (next_val == '*') or \
                               (next_val == ',') or (next_val == ';')
             if (prev_val_check and next_val_check):
                new_char_val = char_val
             else:
                new_char_val = '_'
          expression_inp_cleaned = expression_inp_cleaned + new_char_val
          i = i + 1
       return(expression_inp_cleaned)

   def eval_expression(self,eval_nodes_dict,expression_inp):
       #expression_inp_cleaned = expression_inp.replace('.','_')
       expression_inp_cleaned = self.clean_expression_string(expression_inp)

       eval_functions_dict = {}
       eval_functions_dict.update({'sin': np.sin})
       eval_functions_dict.update({'cos': np.cos})
       eval_functions_dict.update({'min': np.min})
       eval_functions_dict.update({'max': np.max})
       eval_functions_dict.update({'__builtins__': {}})

       try:
          eval_out = eval(expression_inp_cleaned,eval_nodes_dict,eval_functions_dict)
       except (SyntaxError, NameError, TypeError, ZeroDivisionError):
          except_message = traceback.format_exc()

          mssg_index = except_message.find('SyntaxError')
          if (mssg_index >= 0):
             print("Error:  syntax error in command expression:")
             print("--> %s" % except_message[mssg_index:])
          mssg_index = except_message.find('NameError')
          if (mssg_index >= 0):
             print("Error:  unknown variable/function name in command expression:")
             print("--> %s" % except_message[mssg_index:])
          mssg_index = except_message.find('TypeError')
          if (mssg_index >= 0):
             print("Error:  type error in command expression:")
             print("--> %s" % except_message[mssg_index:])
          mssg_index = except_message.find('ZeroDivisionError')
          if (mssg_index >= 0):
             print("Error:  divide by zero error in command expression:")
             print("--> %s" % except_message[mssg_index:])
          eval_out = np.array([])
       return(eval_out)

   def deciplot(self,x_in,y_in):
       num_plot_samples_half = int(10e3)  ## output samples is 2x due to interleaved max/min
       num_plot_samples = 2*num_plot_samples_half
       if len(x_in) != len(y_in):
          print("Warning in 'deciplot():  length of x and y must be the same")
          print("  -> In this case, len(x_in) = %d and len(y_in) = %d" % (len(x_in),len(y_in)))
          return(np.empty(0),np.empty(0))
       num_in_samples = len(x_in)
       if num_in_samples > num_plot_samples:
          num_x_samples_per_plot_sample = num_in_samples//num_plot_samples
          num_y_samples_per_plot_half_sample = 2*num_x_samples_per_plot_sample

          num_plot_samples_half_extended = num_in_samples//(2*num_x_samples_per_plot_sample)
          num_plot_samples_extended = num_plot_samples_half_extended*2
          #print("num_x_samples_per_plot_sample = %d, extended = %d, half_extended = %d" % (num_x_samples_per_plot_sample,num_plot_samples_extended,num_plot_samples_half_extended))


          num_in_samples_pruned = num_x_samples_per_plot_sample*num_plot_samples_extended
          x_mat = x_in[:num_in_samples_pruned].reshape(num_plot_samples_extended,num_x_samples_per_plot_sample)
          x_out = x_mat.mean(axis=1)
    
          y_mat = y_in[:num_in_samples_pruned].reshape(num_plot_samples_half_extended,num_y_samples_per_plot_half_sample)
          y_min = y_mat.min(axis=1)
          y_max = y_mat.max(axis=1)
          y_out = np.empty((num_plot_samples_extended,), dtype=x_in.dtype)
          y_out[0::2] = y_min
          y_out[1::2] = y_max
       else:
          x_out = x_in
          y_out = y_in
       return(x_out,y_out)

   def extract_time(self,t_in):
       Ts = np.mean(np.diff(t_in))
       if self.simulator_name == 'CppSim':
          t_out = Ts*np.arange(len(t_in)) + t_in[0]
       else:
          t_out = t_in
       return(t_out,Ts)

   def apply_zoomx(self,x_in,y_in):
       self.state_multiplot.set_xmin(x_in[0])
       self.state_multiplot.set_xmax(x_in[-1])                
       if self.state_multiplot.state == 'ZoomXall':
          x_min = self.state_multiplot.get_x0()
          x_max = self.state_multiplot.get_x1()
          if x_min == x_max:
             self.state_multiplot.set_x0(x_in[0])
             self.state_multiplot.set_x1(x_in[-1])             
             x_out = x_in
             y_out = y_in
          else:
             x_out = x_in[(x_in >= x_min) & (x_in <= x_max)]
             y_out = y_in[(x_in >= x_min) & (x_in <= x_max)]
             #print("x_min = %5.3e, x_max = %5.3e, x_in[0] = %5.3e, x_in[-1] = %5.3e" % (x_min,x_max,x_in[0],x_in[-1]))
             #print("  ->  x_out[0] = %5.3e,  x_out[-1] = %5.3e" % (x_out[0],x_out[-1]))          
       else:
          x_out = x_in
          y_out = y_in
       return(x_out,y_out)

   def plot_invalid_command(self,t_in,plot_db):
      y_in = np.ones(len(t_in))*np.NaN
      x_plot_label = 'Time'
      xunit_plot_label = 'seconds'
      if plot_db == True:
         y_plot_label = 'dB'
      else:
         y_plot_label = ''      
      yunit_plot_label = ''

      x_out = t_in
      if plot_db == True:
         y_out = 20*np.log10(abs(y_in+1e-30))
      else:
         y_out = y_in
      right_axis_offset = 0
      return(x_out,y_out,x_plot_label,xunit_plot_label,y_plot_label,yunit_plot_label,right_axis_offset)

   #@profile
   def cpp_multiplot(self,fig,plot_node_expression):
       plot_node_expression_no_space = plot_node_expression.replace(' ','')
       eval_nodes_dict = self.extract_nodes(plot_node_expression_no_space)

       subplot_expression_list = plot_node_expression_no_space.split(';')
       num_subplots = len(subplot_expression_list)
       ax_list = []
       subplot_num = 0

       for subplot_expression in subplot_expression_list:
          subplot_num = subplot_num + 1
          plot_expression_list = subplot_expression.split(',')
          ax = fig.add_subplot(num_subplots,1,subplot_num)
          ax_list.append(ax)
          for plot_expression in plot_expression_list:
             #print("%d:  %s" % (subplot_num,plot_expression))
             y_raw = self.eval_expression(eval_nodes_dict,plot_expression)
             #print("******** y_raw ********")
             #print(y_raw)
             [t_raw,Ts] = self.extract_time(self.data.evalsig('TIME'))
             if len(y_raw) > 0: 
                [x_full,y_full,self.x_plot_label,self.xunit_plot_label,self.y_plot_label,self.yunit_plot_label,right_axis_offset] = self.plot_command(t_raw,y_raw,self.plot_db)
             else:
                [x_full,y_full,self.x_plot_label,self.xunit_plot_label,self.y_plot_label,self.yunit_plot_label,right_axis_offset] = self.plot_invalid_command(t_raw,self.plot_db)

             # assume min and max of x-axis are first and last values, respectively
             self.xmin_full = x_full[0]
             self.xmax_full = x_full[-1]
             #print("xmin_full = %5.3e, xmax_full = %5.3e" % (self.xmin_full,self.xmax_full))
             xunit_label,xscale_factor = self.determine_unit_prefix(x_full)
             self.xunit_plot_label_prefix = xunit_label
             self.xunit_scale_factor = xscale_factor 
             self.yunit_plot_label_prefix = ''             
             self.yunit_scale_factor = 1                          
             [x_select,y_select] = self.apply_zoomx(x_full,y_full)
             [x,y] = self.deciplot(x_select*xscale_factor,y_select)

             if y.size != 0:
                if self.plot_logx == False and self.plot_logy == False:
                   ax.plot(x,y,label=plot_expression,linestyle='-',marker=self.plot_marker_style)
                elif self.plot_logx == True and self.plot_logy == False:
                   ax.semilogx(x,y,label=plot_expression,linestyle='-',marker=self.plot_marker_style)
                   if self.plot_command == phase_noise and right_axis_offset != 0.0:
                      y_min,y_max = ax.get_ylim()
                      ax_right = ax.twinx()
                      ax_right.set_ylim(y_min+right_axis_offset,y_max+right_axis_offset)
                      ax_right.set_ylabel('dBc')
                elif self.plot_logx == False and self.plot_logy == True:
                   ax.semilogy(x,y,label=plot_expression,linestyle='-',marker=self.plot_marker_style)
                elif self.plot_logx == True and self.plot_logy == True:
                   ax.loglog(x,y,label=plot_expression,linestyle='-',marker=self.plot_marker_style)
                xlimits = ax.get_xlim()
                if self.plot_xmin == '':
                   xmin_val = xlimits[0]
                else:
                   xmin_val = float(self.plot_xmin)*xscale_factor
                if self.plot_xmax == '':
                   xmax_val = xlimits[1]
                else:
                   xmax_val = float(self.plot_xmax)*xscale_factor
                ax.set_xlim([xmin_val,xmax_val])
                ax.xaxis.set_major_formatter(FormatStrFormatter('%g'))                 
       return(ax_list)

   def load_data(self):
       self.fill_listbox_node_names()

   def eval_text_command(self,event):
       self.plot_text_command()

   def eval_xmin_update(self,event):
       #xmin_val = self.entry_xmin_update.get()
       #print("xmin_val = '%s'" % xmin_val)
       self.modify_listbox_plot_names()
       self.plot_text_command()

   def eval_xmax_update(self,event):
       #xmax_val = self.entry_xmax_update.get()
       #print("xmax_val = '%s'" % xmax_val)
       self.modify_listbox_plot_names()
       self.plot_text_command()

   def update_plot_command_history(self):
      plot_file_name = self.cur_test_tr_filename
      plot_command_name = self.plot_command.__name__
      plot_command_history = self.plot_command_history
      #print("Update:  cur_test_tr_filename = '%s'" % plot_file_name)
      #print("plot_command_name:  '%s'"%plot_command_name)
      if plot_command_history.get_plot_command_name() == plot_command_name and self.plot_command_history.get_plot_file_name() == '':
         #print("---- Setting filename -----")
         plot_command_history.set_plot_file_name(plot_file_name)
      elif plot_command_history.get_plot_command_name() != plot_command_name or self.plot_command_history.get_plot_file_name() != plot_file_name:
         find_command_history_flag = False
         for plot_command_history in self.plot_command_history_list:
            if plot_command_history.get_plot_command_name() == plot_command_name and plot_command_history.get_plot_file_name() == plot_file_name:
               #print("---- Found filename -----")
               self.plot_command_history = plot_command_history
               find_command_history_flag = True
               break
         if find_command_history_flag == False:
            new_plot_command_history = PlotCommandHistory()
            self.plot_command_history_list.append(new_plot_command_history)
            index_val = self.listbox_plot_names.curselection()
            index_val_int = index_val[0]  
            plot_command_arg_list = self.listbox_plot_names.get(index_val_int).split(',')
            #print("------- plot_command_arg_list ------")
            #print(plot_command_arg_list)
            #print("------------------------------------")
            self.init_plot_command_history(new_plot_command_history,plot_command_name,plot_command_arg_list,plot_file_name)
            self.plot_command_history = new_plot_command_history

   #@profile
   def plot_text_command(self):
       command_string = self.text_command.get()
       if len(command_string) > 0:
          self.update_plot_command_history()
          self.plot_command_history.add_entry(command_string)
          #print("-----") 
          #print(event)
          #print(command_string)

          self.fig.clf()
          #self.ax_list = self.cpp_multiplot(self.fig,data,'cos(vin)+sin(sd_in), sin(vin*sd_in)  ;   vin-sd_in, sin(sd_in); vin; sd_in; vin*1e-3; xi12.xor_out')
          self.ax_list = self.cpp_multiplot(self.fig,command_string)
          for ax in self.ax_list:
              if ax != self.ax_list[-1]:
                 ax.xaxis.set_tick_params(labelbottom=False)
              if self.yunit_plot_label == '':
                 ax.set_ylabel("%s" % (self.y_plot_label))
              else:
                 ax.set_ylabel("%s (%s%s)" % (self.y_plot_label,self.yunit_plot_label_prefix,self.yunit_plot_label))
              ax.grid(linestyle='--')
              ax.legend(loc='upper right')
              #print(ax.get_shared_y_axes())
              #if ax != self.ax_list[-1]:
              #   self.ax_list[-1].sharex(ax)
          if self.xunit_plot_label == '':
              self.ax_list[-1].set_xlabel("%s" % (self.x_plot_label))
          else:
              self.ax_list[-1].set_xlabel("%s (%s%s)" % (self.x_plot_label,self.xunit_plot_label_prefix,self.xunit_plot_label))
          if self.plot_command == phase_noise:              
             self.fig.subplots_adjust(left=0.09,right=0.91,top=0.95,bottom=0.1)
          else:
             self.fig.subplots_adjust(left=0.09,right=0.97,top=0.95,bottom=0.1)             
          self.canvas.draw()
       else:
          self.fig.clf()
          self.canvas.draw()

   #def plot_span_selector(self,tmin,tmax):
   #    print(tmin,tmax)

   def plot(self):
       self.plot_text_command()
       
   def load_data_and_replot(self):
       self.load_data()
       self.plot_text_command()

   def synch_to_cur_simruns_dir(self):
       ## Obtain Current Simruns Directory
       self.cur_simruns_path = self.obtain_cur_simruns_path()
       lib_cellname = self.cur_simruns_path.split('SimRuns/',1)[1]
       lib_cellname_list = lib_cellname.split('/',1)
       #print(lib_cellname_list)
       lib_cellname_string = "Library:  %s,  Module:  %s" % (lib_cellname_list[0],lib_cellname_list[1])
       #print("SimRuns Path: '%s'" % self.cur_simruns_path)
       self.title("CppSimPyView (by M.H. Perrott)  ---  " + lib_cellname_string)       
       ## Obtain Par File Names and Add to Test Par Listbox
       cur_simruns_path_filelist = os.listdir(self.cur_simruns_path)
       par_file_names_list = []
       for filename in cur_simruns_path_filelist:
           if '.par' in filename and '.par~' not in filename:
               par_file_names_list.append(filename)
       par_file_names_list.sort()
       self.fill_listbox_test_par_names(par_file_names_list)

   def obtain_cur_simruns_path(self):
       current_cell_sue2_filename = '%s/Sue2/current_cell.cppsimview' % CPPSIM_PATH
       if not os.path.isfile(current_cell_sue2_filename):
           exists_current_cell_sue2_filename = False
           print("Warning:  No Sue2 Current CppSimView Cell File '%s'" % current_cell_sue2_filename) 
           time_current_cell_sue2_filename = -1
       else:
           exists_current_cell_sue2_filename = True
           time_current_cell_sue2_filename = os.path.getmtime(current_cell_sue2_filename)
           
       current_netlist_filename = '%s/Netlist/netlist.cppsim' % CPPSIM_PATH
       if not os.path.isfile(current_netlist_filename):
           exists_current_netlist_filename = False
           print("Warning:  No Netlist File '%s'" % current_netlist_filename) 
           time_current_netlist_filename = -1
       else:
           exists_current_netlist_filename = True
           time_current_netlist_filename = os.path.getmtime(current_netlist_filename)

       cur_module_sue2_name = ''
       if time_current_cell_sue2_filename > time_current_netlist_filename:
           with open(current_cell_sue2_filename) as fp:
               lines = fp.read()
               cur_module_sue2_name = lines.split('\n',1)[0]
               fp.close()
       if len(cur_module_sue2_name) < 1 and time_current_netlist_filename > 0:
           with open(current_netlist_filename) as fp:
               lines = fp.readlines()
               for line in lines:
                   if 'path:' in line:
                      cur_module_sue2_name = line.split(' ',1)[1].split('.sue',1)[0]
                      break
               fp.close()
       if len(cur_module_sue2_name) < 1:
          print("Error:  Cannot Determine Current Module")
          return('')
       lib_cellname = cur_module_sue2_name.split('SueLib/',1)[1]
       cur_simruns_path = "%s/SimRuns/%s" % (CPPSIM_PATH,lib_cellname)
       return(cur_simruns_path)
       
if __name__ == "__main__":
    app = CppSimPyViewApp()
    app.mainloop()
