##### cppsimpyfunctions.py #####
# written by Michael H. Perrott
# available at www.cppsim.com as part of the CppSim package
# Copyright (c) 2004-2023 by Michael H. Perrott
# This file is disributed under the MIT license

# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in
# all copies of this software.

# IN NO EVENT SHALL MICHAEL H PERROTT BE
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
# DOCUMENTATION, EVEN IF MICHAEL H PERROTT
# HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# MICHAEL H PERROTT SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THE SOFTWARE PROVIDED
# HEREUNDER IS ON AN "AS IS" BASIS, AND MICHAEL H PERROTT HAS NO
# OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.

import numpy as np
from scipy.signal import lfilter,welch

def transient(t_in,y_in,plot_db):
   x_plot_label = 'Time'
   xunit_plot_label = 'seconds'
   if plot_db == True:
      y_plot_label = 'dB'
   else:
      y_plot_label = ''      
   yunit_plot_label = ''

   x_out = t_in
   if plot_db == True:
      y_out = 20*np.log10(abs(y_in))
   else:
      y_out = y_in
   right_axis_offset = 0      
   return(x_out,y_out,x_plot_label,xunit_plot_label,y_plot_label,yunit_plot_label,right_axis_offset)

def spectrum(t_in,y_in,plot_db):
   x_plot_label = 'Frequency'
   xunit_plot_label = 'Hertz'
   if plot_db == True:
      y_plot_label = 'dB/Hz'
   else:
      y_plot_label = 'V/rHz'      
   yunit_plot_label = ''       

   Ts = np.mean(np.diff(t_in))   
   f, Pxx = welch(y_in,fs=1/Ts,window='hann',nperseg=2**16,noverlap=None,nfft=None,detrend='constant',return_onesided=True,scaling='density',axis=-1,average='mean')
   x_out = f
   if plot_db == True:
      y_out = 10*np.log10(np.abs(Pxx))
   else:
      y_out = np.sqrt(np.abs(Pxx))
   right_axis_offset = 0            
   return(x_out,y_out,x_plot_label,xunit_plot_label,y_plot_label,yunit_plot_label,right_axis_offset)

def phase_noise(t_in,y_in,plot_db):
   x_plot_label = 'Frequency'
   xunit_plot_label = 'Hertz'
   if plot_db == True:
      y_plot_label = 'dBc/Hz'
   else:
      y_plot_label = 'Radians/rHz'      
   yunit_plot_label = ''       

   Ts = np.mean(np.diff(t_in))
   Kv = 1.0
   phase = lfilter([Ts*2*np.pi*Kv],[1,-1],y_in-np.mean(y_in))
   f, Pxx = welch(phase,fs=1/Ts,window='hann',nperseg=2**16,noverlap=None,nfft=None,detrend='constant',return_onesided=True,scaling='density',axis=-1,average='mean')
   x_out = f
   if plot_db == True:
      y_out = 10*np.log10(np.abs(0.5*Pxx))
   else:
      y_out = np.sqrt(np.abs(0.5*Pxx))
   right_axis_offset = 0            
   return(x_out,y_out,x_plot_label,xunit_plot_label,y_plot_label,yunit_plot_label,right_axis_offset)
